#include <stddef.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
//#include <stdio.h>
//#include <string.h>

#include <avr/io.h>
#include <avr/interrupt.h>
//#include <avr/pgmspace.h>
//#include <avr/sleep.h>
//#include <avr/eeprom.h>
//#include <avr/wdt.h>

#include <util/delay.h>

#include "uart.h"

#define CMD_SYNC		0
#define CMD_RESET		1
#define CMD_GET_TIME	2
#define CMD_GET_FCPU	3

#define TIMER1_DELTA_COUNT	(F_CPU/64/1000)		// 1000 Hz

volatile register uint16_t msec asm("r2");	// r2, r3
volatile register uint32_t sec asm("r4");	// r4, r5, r6, r7

// 1000 Hz
ISR(TIMER1_COMPA_vect) {
	OCR1A += TIMER1_DELTA_COUNT;

	msec++;
	if (msec >= 1000) {
		msec = 0;
		sec++;
	}
}


static void uartInit() {
	#if UART_DOUBLE_SPEED
		#if UART0_BIT_U2X
			UART0_STATUS = _BV(UART0_BIT_U2X);  // Enable 2x speed
		#endif
	#else
		UART0_STATUS = 0;  					// Disable 2x speed
	#endif


	#if defined(UART0_UBRRH)
		UART0_UBRRH = (uint8_t) ((UART_UBRR >> 8) & 0xFF) ;
	#endif
	UART0_UBRRL = (uint8_t) (UART_UBRR & 0xFF);

	// Enable USART receiver and transmitter
	UART0_CONTROL = _BV(UART0_BIT_RXEN) | _BV(UART0_BIT_TXEN);

	// Set frame format: asynchronous, 8data, no parity, 1stop bit
	#ifdef UART0_CONTROLC
		#ifdef UART0_BIT_URSEL
			UART0_CONTROLC = _BV(UART0_BIT_URSEL) | _BV(UART0_BIT_UCSZ1) | _BV(UART0_BIT_UCSZ0);
		#else
			UART0_CONTROLC = _BV(UART0_BIT_UCSZ1) | _BV(UART0_BIT_UCSZ0);
		#endif
	#endif
}


static void initTimer() {
	// TC2
	TCCR1A = 0;
	TCCR1B = _BV(CS11) | _BV(CS10);	// divider for TC1 = 64
	TIMSK = _BV(OCIE1A);		
	
	sei();
}


static void uartPutChar(uint8_t data) {
	while (!(UART_STATUS & _BV(UART_TXREADY)));
	UART_DATA = data;
}

static uint8_t uartGetChar() {
	while (!(UART_STATUS & _BV(UART_RXREADY)));
	return UART_DATA;
}

static void uartPutWord(uint16_t w) {
	uartPutChar(w >> 8);
	uartPutChar(w & 0xFF);
}

static void uartPutDword(uint32_t dw) {
	uartPutChar(dw >> 24);
	uartPutChar((dw >> 16) & 0xFF);
	uartPutChar((dw >> 8) & 0xFF);
	uartPutChar(dw & 0xFF);
}

static void processCommand(uint8_t cmd) {
	switch(cmd) {
		case CMD_SYNC:
			uartPutChar(uartGetChar());
			break;
		case CMD_RESET:
			cli();
			TCNT1 = 0;
			msec = 0;
			sec = 0;
			sei();
			break;
		case CMD_GET_TIME:
			uartPutWord(msec);
			uartPutDword(sec);
			break;
		case CMD_GET_FCPU:
			uartPutDword(F_CPU);
			break;
	}
}



void main(void) {
	uartInit();
	initTimer();
	while (true) {
		processCommand(uartGetChar());
	}
}
